import importlib.util
import os
from senpai_lang.src.parser import parser
import logging
import pathlib
logger = logging.getLogger()

class SenpaiModule:
	"""The base class that imports a Senpai module. Imports from the relative or absolute file path."""
	lib_path = pathlib.Path(os.path.split(os.path.realpath(__file__))[0]).parent / 'lib'

	def __init__(self, code):
		self.code = code
		
	def run(self, interpreter_class):
		
		interpreter = interpreter_class(parser.parse(self.code), python_traceback=True)
		return interpreter.vars

	@classmethod
	def from_path(cls, path):
		logger.info(f"Trying to import senpai module from {path!r}")
		path = path.replace('%lib%', str(cls.lib_path))
		with open(path, "r") as f:
			code = f.read()
		module = cls(code)
		name = os.path.splitext(os.path.basename(path))[0]
		module.name = name
		return module
	
class PythonModule:
	"""The base class that imports a Senpai Python module to be accessable from the Senpai Interpreter. Imports from the relative or absolute file path."""

	lib_path = pathlib.Path(os.path.split(os.path.realpath(__file__))[0]).parent / 'lib'

	def __init__(self, module, spec, name):
		self.module = module
		self.spec = spec
		self.name = name
	
	@classmethod
	def from_path(cls, path):
		path = path.replace('%lib%', str(cls.lib_path))
		spec = importlib.util.spec_from_file_location('', path)
		module = importlib.util.module_from_spec(spec)
		return PythonModule(module, spec, 'module')

	def run(self, interpreter_class):
		self.spec.loader.exec_module(self.module)
		





from senpai_lang.pyutils import FuncBase, PyModule
import os


class System(FuncBase):
	"""
	Can take 2 or 1 argument
	First argument is always the command to execute.
	Second argument can be anything, 
	
	Second argument if it exists returns the value os.system()
	
	On Unix, the return value is the exit status of the process encoded in the format specified for wait().

	On Windows, the return value is that returned by the system shell after running command. The shell is given by the Windows environment variable COMSPEC: it is usually cmd.exe, which returns the exit status of the command run; on systems using a non-native shell, consult your shell documentation.
	"""
	name = 'cmd'
	def __init__(self):
		super().__init__(self.name, self.system)

	@staticmethod
	def system(*args):
		if len(args) == 1:
			os.system(*args)
		else:
			return os.system(args[0])
			
PyModule('os', [System])


from senpai_lang.src.classes import SenpaiInterpreter
import time
from senpai_lang.src.parser import parser
import logging
import argparse
import senpai_lang.src.senpai_builtins
# import os

cmdparse = argparse.ArgumentParser()
cmdparse.add_argument("-i", "--infile", type=str, help="Input file for interpreter") # for some reason using argparse.FileType and then reading causes EOF error with Lark. Idk why.
cmdparse.add_argument("-d", "--debugfile", default=None, type=str, help="File to write to for verbose logging. The log tracks almost everything the intepreter does down to the expression.")

cmdparse.add_argument("-tb", "--pytraceback", default=False, action='store_true', help="Enable Python exception traceback. Useful for taking note of bugs. Disables the Senpai except hook.")

cmdparse.add_argument("-ps", "--printstack", default=False, action='store_true', help="Print the execution time, variables, stacks, and builtins of the program after running")

cmdparse.add_argument("-w", "--wiki", default=False, action='store_true', help="Visit the wiki page for the Senpai language")

def main():
	args = cmdparse.parse_args()

	if args.wiki:
		print("https://esolangs.org/wiki/Senpai")
		exit(0)
		
	if args.debugfile:
		logging.basicConfig(filename=args.debugfile, level=logging.DEBUG, format='%(asctime)s.%(msecs)03d - %(message)s',datefmt='%Y-%b-%d, %H:%M:%S', )

		logging.info("\n\n\n===Starting logging of Senpai===")

	with open(args.infile) as f:
		# os.chdir(os.path.split(os.path.realpath(args.infile))[0])
		t = f.read()
	tree = parser.parse(t)
	# print(tree)
	p = time.perf_counter()
	i = SenpaiInterpreter(tree, python_traceback=args.pytraceback)

	t = (time.perf_counter() - p)
	logger = logging.getLogger()
	logger.info(f"Program executed successfully in {t} seconds!")
	if args.printstack:
		print("Time taken:", t*1000, "miliseconds")
		print("Variables:", i.vars)
		print("Stacks:", i.stacks)
		print("Builtins:", i.builtins)

if __name__ == "__main__":
	main()



from senpai_lang.pyutils import FuncBase, PyModule
import os
class Love(FuncBase):
    """Wrapper for python's print() function"""
    name = 'love'
    def __init__(self):
        super().__init__(self.name, print)
    
class Reason(FuncBase):
    """Wrapper for python input() function"""
    name = 'reason'
    def __init__(self):
        super().__init__(self.name, input)

class PrintStack(FuncBase):
    """Prints the current stack. Meant for debugging."""
    name = 'stack'
    def __init__(self):
        super().__init__(self.name, self.print_stack)

    def print_stack(self):
        print(self.interpreter.stacks.current_stack)

class Crash(FuncBase):
    """
    Exits the underlying python process. Wraps python exit() function
    """
    name = 'crash'
    def __init__(self):
        super().__init__(self.name, os._exit)
# You can manipulate the stacks of the interpreter with self.interpreter.stacks 
# or even the vars with self.interpreter.vars
        
PyModule("__builtins__", [Love, Reason, PrintStack, Crash], rename_funcs=False)


import logging
from senpai_lang.src.classes import SenpaiBuiltinFunction, SenpaiInterpreter
logger = logging.getLogger()


FuncBase = SenpaiBuiltinFunction

class PyModule:
	def __init__(self, name, functions, *, rename_funcs=True):
		logger.info(f"Importing Python module {name!r}")
		self.name = name
		self.functions = functions
		self.rename = rename_funcs
		if self.rename:
			
			for func in self.functions:
				func.name = f'{self.name}_{func.name}'
		logger.info(f"Adding functions {[i.name for i in self.functions]}")
		for func in self.functions:
			func()
		logger.info(f"Imported {self.name!r} successfully")
		logger.info(f"State of builtins: {SenpaiInterpreter.builtins}")


"""
Extensions of interpreter's stack manipulation through functions. Allows creation and deletion of stacks with names that are determined by variables. Useful to use with list module, as the Stack class and the SenpaiList class both inherit off of collections.deque
"""
from senpai_lang.pyutils import FuncBase, PyModule

class CurrentStack(FuncBase):
	"""Returns a reference to the current stack. Things can get weird..."""
	name = "cur"
	def __init__(self):
		super().__init__(self.name, self.current_stack)

	def current_stack(self):
		return self.interpreter.stacks.current_stack

class GetStack(FuncBase):
	"""Returns a reference to the stack with that name. All changes to that stack will be reflected in the stack returned and vice versa."""
	name = "get"
	def __init__(self):
		super().__init__(self.name, self.get_stack)

	def get_stack(self, name):
		return self.interpreter.stacks[name]

class SwitchStack(FuncBase):
	"""Switch the current stack to a stack with that name. Creates an empty stack if the name does not exist. Essentially the same thing as the builtin instruction, but it can take any hashable object as the key and allows the creation of stacks with undetermined names"""
	name = "switch"
	def __init__(self):
		super().__init__(self.name, self.switch_stack)

	def switch_stack(self, name):
		self.interpreter.stacks.switch_stack(name)

class DeleteStack(FuncBase):
	"""Deletes a stack with a given name"""
	name = "del"
	def __init__(self):
		super().__init__(self.name, self.del_stack)

	def del_stack(self, name):
		del self.interpreter.stacks[name]

PyModule('stack', [CurrentStack, GetStack, SwitchStack, DeleteStack])
	



import decimal
import lark
import operator
from collections import deque
from senpai_lang.src.senpai_import import SenpaiModule, PythonModule
import sys
import logging


class SenpaiInterpreter(lark.visitors.Interpreter):	
	builtins = {}
	def __init__(self, tree, *, python_traceback=False):
		self.id = hex(id(self))
		logging.info(f"Initializing Interpreter at {self.id}")
		if not python_traceback:
			sys.excepthook = self.custom_except_hook
			logging.info(self.format_log(f"Created custom except hook"))

		logging.info(self.format_log(f"Interpreting the following tree:\n {tree.pretty()}"))
		self.tree = tree
		self.vars = {}
		self.stacks = StackHolder('bedroom', **{'bedroom': Stack()})
		self.visit(tree)
	
	def custom_except_hook(self, _type, value, tb):
		print("Sorry senpai, you messed up somewhere UmU", file=sys.stderr)
		
		print(f"{_type.__name__}: {value}", file=sys.stderr, end='\n\n')
		print("The status of stacks before the error occured were:\n", file=sys.stderr)

		for name in self.stacks:
			print(f'\t{name}: ', file=sys.stderr, end='\n\t\t')
			for i, item in enumerate([repr(item) for item in self.stacks[name]]):
				print(i, item, file=sys.stderr,sep='\t', end='\n\t\t')
			print()

		print("\nThe values of all of the variables before the error were:", file=sys.stderr, end='\n\n\t')

		for var in self.vars:
			print(f"{var}:", repr(self.vars[var]), file=sys.stderr, end='\n\t')
		print('\b \b', file=sys.stderr)
		logging.error(f"Exception was raised: {_type.__name__}: {value}")

	def format_log(self, message):
		return f"\n{self.id} - {message}\n"

	def declaration(self, tree):
		return Declaration(tree, self)	
	
	def assignment(self, tree):
		return Assignment(tree, self)

	def stack_op(self, tree):
		op = tree.children[0]
		op_name = op.data

		if op_name == 'push':
			var_name = op.children[0].children[0].value
			logging.info(self.format_log(f"Pushing {var_name!r} onto the current stack"))
			try:
				self.stacks.current_stack.appendleft(self.vars[var_name])
			except KeyError:
				self.stacks.current_stack.appendleft(SenpaiInterpreter.builtins[var_name])
		elif op_name == 'pop':
			logging.info(self.format_log("Popping TOS"))
			self.stacks.current_stack.popleft()
		elif op_name == 'rot2':
			logging.info(self.format_log("Swapping top 2 items on the stack"))
			self.stacks.current_stack.rot2()
		elif op_name == 'rot3':
			logging.info(self.format_log("Swapping top 3 items on the stack"))
			self.stacks.current_stack.rot3()
		elif op_name == 'switch_stack':
			
			stack_name = op.children[0].children[0].value
			logging.info(self.format_log(f"Switching the current stack to {stack_name!r}"))
			self.stacks.switch_stack(stack_name)
		d = {self.stacks.current_stack_name: self.stacks.current_stack}
		logging.info(self.format_log(f"State of current stack: {d}"))
		
	def if_block(self, tree):
		IfBlock(tree, self).run()

	def if_else_block(self, tree):
		IfElseBlock(tree, self).run()
		
	def loop(self, tree):
		Loop(tree, self).run()
	
	def call(self, tree):
		number_args = len(tree.children[0].value) if tree.children else 0
		stack = self.stacks.current_stack

		if (number_args + 1) > len(stack):
			func = self.stacks.current_stack[0]
			raise ValueError(f"Insufficient number of arguments on the stack. Failed to call function {func.name!r}.")

		function = self.stacks.current_stack.popleft()

		d = {self.stacks.current_stack_name: self.stacks.current_stack}

		logging.info(self.format_log(f"Calling function {function} with {number_args} args"))
		logging.info(self.format_log(f"Current stack state before calling function: {d}"))
		if number_args:
			args = [self.stacks.current_stack.popleft() for _ in range(number_args)] 
		else:
			args = []
		function.call(args, self)

		logging.info(self.format_log(f"Current stack state after calling function: {d}"))
	def function_definition(self, tree):
		return SenpaiFunction(tree, self)
	
	def import_statement(self, tree):
		path = tree.children[0].children[0].value.strip('"')
		logging.info(self.format_log(f"Importing module from {path!r}"))
		try:
			module = SenpaiModule.from_path(path)
			self.vars.update(module.run(self.__class__))
			logging.info(self.format_log(f"Import of Senpai module at {path!r} was successful"))

		except lark.exceptions.UnexpectedCharacters:
			try: 
				PythonModule.from_path(path).run(SenpaiInterpreter)
				logging.info(self.format_log(f"Import of Python module at {path!r} was successful"))
			except AttributeError:
				raise ImportError(f"Could not find a module at {path!r}")
	
	def transfer_top(self, tree):
		d = {self.stacks.current_stack_name: self.stacks.current_stack}

		item = self.stacks.current_stack.popleft()
		var_name = tree.children[0].children[0].value

		logging.info(self.format_log(f"Transferring TOS ({item}) to variable {var_name!r}"))
		logging.info(self.format_log(f"Current interpreter's stack status before transfer: {d}"))
		logging.info(self.format_log(f"Current interpreter's var status before transfer: {self.vars}"))

		if var_name not in self.vars:
			raise NameError(f"{var_name!r} is not defined!")
		self.vars[var_name] = item
	
	def delete_var(self, tree):
		var_name = tree.children[0].children[0].value
		logging.info(self.format_log(f"Deleting variable {var_name!r}"))
		logging.info(self.format_log(f"Current interpreter's var status before deletion: {self.vars}"))
		del self.vars[var_name]
		logging.info(self.format_log(f"Current interpreter's var status after deletion: {self.vars}"))
		logging.info(self.format_log(f"Deleted {var_name!r} successfully!"))

class Declaration:
	def __init__(self, tree, interpreter):
		assert len(tree.children) == 1
		self.interpreter = interpreter
		self.var_name = tree.children[0].children[0].value
		logging.info(interpreter.format_log(f"Processing declaration of variable {self.var_name!r} from the following tree:\n {tree.pretty()}"))
		
		if self.var_name in self.interpreter.vars:
			raise NameError(f"{self.var_name!r} is already declared!")

		self.interpreter.vars[self.var_name] = None
		logging.info(interpreter.format_log(f"State of variables: {interpreter.vars!r}"))


class Assignment:
	def __init__(self, tree, interpreter):
		children = tree.children
		self.interpreter = interpreter
		self.var_name = children[0].children[0]

		logging.info(interpreter.format_log(f"Processing assignment of variable {self.var_name.value!r} from the following tree:\n{tree.pretty()}"))

		if self.var_name not in self.interpreter.vars:
			raise NameError(f"{self.var_name!r} is not declared!")
		
		expression_value = Expression(tree.children[1], self.interpreter).parse().value[0]
		
		
		logging.info(interpreter.format_log(f"Assigning {self.var_name.value!r} to {expression_value!r}"))

		self.interpreter.vars[self.var_name] = expression_value

		logging.info(interpreter.format_log(f"Assignment to {self.var_name.value!r} succeeded!"))

class Expression:
	binop_dict = {
				'add': operator.add,
				'sub': operator.sub,
				'mul': operator.mul,
				'div': operator.truediv,
				'bit_or': operator.__or__,
				'xor': operator.xor,
				'bit_and': operator.__and__,
				'equality': operator.eq,
				'less_than': operator.lt,
				'greater_than': operator.gt,
				'greater_than_or_equals': operator.ge,
				'less_than_or_equals': operator.le,
				'logical_or': lambda a, b: a or b,
				'logical_and': lambda a, b: a and b,
				'modulus':  operator.mod,
				'inequality': operator.ne
			}
	unary_dict = {
		'neg': operator.neg,
		'bit_not': operator.__invert__,
		'char_convert': chr
	}
	def __init__(self, tree, interpreter):
		self.value = []
		self.tree = tree
		self.interpreter = interpreter
		logging.info(interpreter.format_log(f"Parsing expression from the following tree \n{tree.pretty()}"))

	@classmethod
	def number(cls, number):
		try: 
			return int(number)
		except ValueError:
			return decimal.Decimal(number)

	def parse(self):
		gen = (child for child in self.tree.children)
		for tree in gen:
			op = tree.data
		
			if op == 'number':
				self.value.append(Expression.number(tree.children[0].value))
			elif op == 'string':
				self.value.append(tree.children[0].value[1:-1])
			elif op == 'name':
				try:
					self.value.append(self.interpreter.vars[tree.children[0].value])
				except KeyError:
					self.value.append(self.interpreter.builtins[tree.children[0].value])
			elif op in Expression.binop_dict:
				self.value.append(Expression.binop_dict[op](*Expression(tree, self.interpreter).parse().value))
			elif op in Expression.unary_dict:
				self.value.append(Expression.unary_dict[op](*Expression(tree, self.interpreter).parse().value,))
		logging.info(self.interpreter.format_log(f"Parsed expression! Returned {self.value[0]}"))
		return self

	def __bool__(self):
		return bool(self.value[0])

class IfBlock:
	def __init__(self, tree, interpreter):
		logging.info(interpreter.format_log(f"Creating conditional from the following tree:\n{tree.pretty()}"))
		self.interpreter = interpreter
		self.truth_value = tree.children[0].data == 'true'
		self.truth_expression = tree.children[1]
		self.body = tree.children[2:]


	def run(self):
		logging.info(self.interpreter.format_log("Parsing if block"))
		if bool(Expression(self.truth_expression, self.interpreter).parse()) == self.truth_value:
			logging.info(self.interpreter.format_log("Truthiness matched"))
			for instruction in self.body:
				self.interpreter.visit(instruction)
		logging.info(self.interpreter.format_log("Finished parsing if block"))

		
class IfElseBlock(IfBlock):

	def run(self):
		logging.info(self.interpreter.format_log("Parsing if else block"))

		if bool(Expression(self.truth_expression, self.interpreter).parse()) == self.truth_value:
			logging.info(self.interpreter.format_log("Truthiness matched!"))
			for instruction in [instruction for instruction in self.body if instruction.data != 'else']:
				self.interpreter.visit(instruction)
		else:
			logging.info(self.interpreter.format_log("Truthiness did not match! Executing other instructions"))
			for instruction in [instruction for instruction in self.body if instruction.data == 'else'][0].children:

				self.interpreter.visit(instruction)
		logging.info(self.interpreter.format_log(f"Finished parsing if else block"))
class Loop(IfBlock):
	def run(self):
		logging.info(self.interpreter.format_log("Parsing loop!"))
		while bool(Expression(self.truth_expression, self.interpreter).parse()) == self.truth_value:
			logging.info(self.interpreter.format_log("Truthiness matched!"))
			for instruction in self.body:# [instruction for instruction in self.body if instruction.data != 'else']:
				self.interpreter.visit(instruction)
			logging.info(self.interpreter.format_log("Looping again"))

class SenpaiBuiltinFunction:
	"""The base class for the Senpai Python function wrapper. Wraps an arbitrary function with position only arguments to be usable in senpai. Pops all items passed in as arguments from the stack. iter_results makes it so that if an iterable that contains results is returned, the results will be put onto the stack in the order they are returned.
	
	You can interact with the stacks of the interpreter by using self.interpreter.stacks, and the current stacks with self.interpreter.stacks.current_stack. 
	
	You can also interact with the variables defined with self.interpreter.vars and all the methods of the lark.visitors.Interpreter class if you want to. Not sure why you would, but you technically can. You can do as much as you want with it. You can even fuck up the AST if you want.
	"""
	def __init__(self, name, function, *,iter_results=False):

		self.name = name
		self.function = function
		self.iter_results = iter_results
		SenpaiInterpreter.builtins[self.name] = self
		# logging.info(f"Current state of builtins: {SenpaiInterpreter.builtins}")
		
	def call(self, args, interpreter):
		# Functions do not have return values. They only put values onto the stack
		logging.info(interpreter.format_log(f"Calling Python function {self.name!r} with arguments of {args}"))
		
		self.interpreter = interpreter
		result = self.function(*args)
		if result is not None: 
			if self.iter_results:
				
				b = [i for i in result][::-1]
				for value in b:
					interpreter.stacks.current_stack.appendleft(value)
			else:
				interpreter.stacks.current_stack.appendleft(result)
	
	def __repr__(self):
		return (f"<Builtin: {self.__class__.__name__}"+'>')

class SenpaiFunction:
	def __init__(self, tree, interpreter):
		logging.info(interpreter.format_log(f"Creating function from the following tree:\n{tree.pretty()}"))

		self.name = tree.children[0].children[0].value
		args_tree = tree.children[1]
		self.interpreter = interpreter
		self.args = [i.children[0].value for i in args_tree.children] if args_tree.data == 'function_args' else None
		self.body = tree.children[2] if self.args else tree.children[1]

		logging.info(interpreter.format_log(f"Adding function with name {self.name!r} to vars"))

		self.interpreter.vars[self.name] = self

		logging.info(interpreter.format_log(f"Current state of vars: {interpreter.vars}"))
		

	def call(self, args, interpreter):
		logging.info(interpreter.format_log(f"Calling Senpai function {self.name!r} with arguments of {args}"))
		if len(args) != len(self.args):
			raise ValueError(f"Expected {len(self.args)} parameters for calling function {self.name!r}")

		logging.info(interpreter.format_log(f"Assigning temporary variables {self.args} to {args}"))
		for i, j in zip(self.args, args):
			interpreter.vars[i] = j
		
		logging.info(interpreter.format_log(f"State of current interpreter's vars: {interpreter.vars}"))
		logging.info(interpreter.format_log(f"Excuting body of function {self.name!r}"))	
		interpreter.visit(self.body)
		logging.info(interpreter.format_log(f"Deleting temporary variables {self.args}"))
		
		for i in self.args:
			del interpreter.vars[i]
		logging.info(interpreter.format_log(f"Function {self.name!r} called successfully"))

class Stack(deque):
	
	def rot2(self):
		self[0], self[1] = self[1], self[0]
	
	def rot3(self):
		self[0], self[1], self[2] = self[2], self[0], self[1]

class StackHolder(dict):
	def __init__(self, current_stack, **kwargs):
		super().__init__(kwargs)
		self.current_stack_name = current_stack
		self.current_stack = self[self.current_stack_name]
		
	def switch_stack(self, name):
		self.current_stack_name = name
		if name in self:
			self.current_stack = self[name]
		else:
			self[name] = Stack()
			self.current_stack = self[name]

	


from senpai_lang.pyutils import FuncBase, PyModule
from pathlib import Path

class OpenFile(FuncBase):
	name = 'open'
	def __init__(self):
		super().__init__(self.name, self.file_setup)

	@staticmethod
	def file_setup(path):
		file = Path(path)
		file.touch(exist_ok=True)
		return open(path, 'rb+')

class ReadFile(FuncBase):
	name = 'read'
	def __init__(self):
		super().__init__(self.name, self.read, iter_results=True)
	
	@staticmethod
	def read(*args):
		if len(args) == 1:
			r = args[0].read(), args[0]
			return r
		elif len(args) == 2:
			print(args)
			return args[1].read(args[0]), args[1]
		else:
			raise IndexError(f"'file_read' takes 1 or 2 arguments, but {len(args)} were given")

class WriteFile(FuncBase):
	name = 'write'
	def __init__(self):
		super().__init__(self.name, self.write)
	
	@staticmethod
	def write(arg, file): 
		file.write(arg.encode() if isinstance(arg, str) else bytes(arg))
		return file

class FileSeek(FuncBase):
	name = 'seek'
	def __init__(self):
		super().__init__(self.name, self.seek)
	
	@staticmethod
	def seek(arg, file):
		file.seek(0, 2) if arg == 'end' else file.seek(arg)
		return file

class FileClose(FuncBase):
	name = 'close'
	def __init__(self):
		super().__init__(self.name, self.close)

	@staticmethod
	def close(file):
		file.close()

class FileTell(FuncBase):
	name = 'tell'
	def __init__(self):
		super().__init__(self.name, self.tell, iter_results=True)

	@staticmethod
	def tell(file):
		return file.tell(), file

PyModule('file', [OpenFile, ReadFile, WriteFile, FileSeek, FileClose, FileTell])




from senpai_lang.pyutils import FuncBase, PyModule
from collections import deque

class SenpaiList(deque):
	def __repr__(self):
		# hide the name deque from users so they don't get scared
		return repr(list(self))

class NewList(FuncBase):
	name = "new"
	def __init__(self):
		super().__init__(self.name, self.new_list)
	
	@staticmethod
	def new_list(*args):
		return SenpaiList(args)

class InsertEnd(FuncBase):
	name = "insert_back"

	def __init__(self):
		super().__init__(self.name, self.insert_end)

	@staticmethod
	def insert_end(value, li):
		li.append(value)
		return li

class InsertBegin(FuncBase):
	name = "insert_front"

	def __init__(self):
		super().__init__(self.name, self.insert_front)

	@staticmethod
	def insert_front(value, li):
		li.appendleft(value)
		return li


# General container stuff
class GetItem(FuncBase):
	name = "get"
	def __init__(self):
		super().__init__(self.name, self.get_item, iter_results=True)

	@staticmethod
	def get_item(index, container):
		return container.__getitem__(index), container

class SetItem(FuncBase):
	name = "set"
	def __init__(self):
		super().__init__(self.name, self.set_item, iter_results=True)

	@staticmethod
	def set_item(obj, index, container):
		return container.__setitem__(index, obj), container

class DeleteItem(FuncBase):
	name = "del"
	def __init__(self):
		super().__init__(self.name, self.del_item, iter_results=True)

	@staticmethod
	def del_item(index, container):
		container.__delitem__(index)
		return container

class Unpack(FuncBase):
	name = "unpack"
	def __init__(self):
		super().__init__(self.name, lambda container: container, iter_results=True)
		
PyModule('list', [NewList, InsertEnd, InsertBegin], rename_funcs=True)
PyModule('container', [GetItem, SetItem, DeleteItem, Unpack], rename_funcs=True)

